"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DataLoader = require("dataloader");
const Debug = require("debug");
const debug = Debug('LayeredCacheLoader');
class EmptyLayer {
    async batchGet(keys) {
        return keys.map(key => new Error(`${key}: value not found.`));
    }
    batchSet(map) {
    }
}
class LayeredCacheLoader {
    constructor() {
        this._loaderChain = [];
        this._finalLoader = new DataLoader(async (keys) => {
            throw new Error(`Failed load keys: ${keys}. A final loader should set.`);
        });
        this.use(new EmptyLayer());
    }
    get _loader() {
        return this._loaderChain[0].loader;
    }
    use(layer) {
        const layerName = `${layer.constructor.name}_${this._loaderChain.length}`;
        debug('Use layer:', layerName);
        const currentLoader = {
            loader: new DataLoader(async (keys) => {
                currentLoader.loader.clearAll();
                debug(`[${layerName}]Begin batch get keys:`, keys);
                try {
                    const results = await layer.batchGet(keys);
                    const newCacheMap = new Map();
                    const finalResults = await Promise.all(results.map(async (result, index) => {
                        if (result instanceof Error) {
                            const key = keys[index];
                            try {
                                const value = await (currentLoader.nextLoader || this._finalLoader).load(key);
                                newCacheMap.set(key, value);
                                return value;
                            }
                            catch (e) {
                                throw e;
                            }
                        }
                        return result;
                    }));
                    if (newCacheMap.size > 0) {
                        debug(`[${layerName}]Cache keys from next layer:`, newCacheMap.keys());
                        layer.batchSet(newCacheMap);
                    }
                    return finalResults;
                }
                catch (e) {
                    throw e;
                }
            }),
        };
        const prevLoader = this._loaderChain[this._loaderChain.length - 1];
        if (prevLoader) {
            prevLoader.nextLoader = currentLoader.loader;
        }
        this._loaderChain.push(currentLoader);
        return this;
    }
    final(batchLoadFn) {
        const finalLoader = new DataLoader(async (keys) => {
            finalLoader.clearAll();
            debug(`[FinalLoader]Begin batch get keys:`, keys);
            return batchLoadFn(keys);
        });
        this._finalLoader = finalLoader;
        return this;
    }
    load(key) {
        return this._loader.load(key);
    }
    loadMany(keys) {
        return this._loader.loadMany(keys);
    }
}
exports.default = LayeredCacheLoader;
//# sourceMappingURL=index.js.map